int nbrBoids = 300;

Boid[] boids = new Boid[nbrBoids];

void setup() 
{ 
  size(700, 500);
  frameRate(30);
  
  for(int i=0; i<nbrBoids; i++)
  {
    boids[i] = new Boid();
  }
}

void draw() 
{
  background(255);

  for(int i=0; i<nbrBoids; i++)
  {
    ellipse(boids[i].pos.x, boids[i].pos.y, boids[i].siz, boids[i].siz); // ball
    boids[i].updatePosition(boids);
  }
} 

class Boid
{
  float cohesion;
  // goToCenter rule intensity
  // .001 : boids regroup according to other rules
  // 1.00 : boids piled ignoring other rules

  float avoidance;
  // keepDistance rule intensity
  // .001 : fluid, boids collision may appear
  // 1.00 : precise and excited, without collisions

  float imitation;
  // matchVelocity rule intensity
  // .001 : each boid do it's own life
  // 1.00 : boids moves like one boid (boid army)

  float siz; // boid size
  
  float aura;
  // keepDistance rule condition
  // each boid try to keep this minimum distance between others
  
  float observation;
  // goToCenter and matchVelocity rules condition
  // max distance for which each boid sees and responds to others
  
  PVector pos;
  PVector vel;

  Boid()
  {
    cohesion = .005;
    avoidance = .1;
    imitation = .08;
    siz = 5;
    aura = 2 * siz;
    observation = 20 * siz;

    pos = new PVector(random( width * 0.1, width * 0.9), random( height * 0.1,  height * 0.9));
    vel = new PVector();
  }
  float distanceTo(Boid other)
  {
    return sqrt( sq(pos.x - other.pos.x) + sq(pos.y - other.pos.y) + sq(pos.z - other.pos.z) );
  }

  PVector goToCenter(Boid[] boids)
  {
    PVector v = new PVector();
    int nbr = 0;
    for(int i=0; i<boids.length; i++)
    {
      if((boids[i] != this) && (distanceTo(boids[i]) < observation)){
        v.add(boids[i].pos);
        nbr +=1;
      }
    }
    if(nbr == 0) return new PVector();
    return PVector.mult(PVector.sub(PVector.div(v, nbr), pos), cohesion);
  }

  PVector keepDistance(Boid[]  boids)
  {
    PVector v = new PVector();
    for(int i=0; i<boids.length; i++)
    {
      if((boids[i] != this) && (distanceTo(boids[i]) < aura))
      {
        v.sub(PVector.sub(boids[i].pos, pos));
      }
    }
    return PVector.mult(v, avoidance);
  }

  PVector matchVelocity(Boid[] boids)
  {
    PVector v = new PVector();
    int nbr = 0;
    for(int i=0; i<boids.length; i++)
    { 
      if((boids[i] != this) && (distanceTo(boids[i]) < observation))
      {
        v.add(boids[i].vel);
        nbr +=1;
      }
    }
    if(nbr == 0) return new PVector();
    return PVector.mult(PVector.sub(PVector.div(v, nbr), vel), imitation);
  }

  void limitVelocity(float lim)
  {
    if(vel.mag() > lim)
    {
      vel = PVector.mult(PVector.div(vel, vel.mag()), lim);
    }
  }

  PVector tendToPlace(int x, int y)
  {
    PVector place = new PVector(x, y);
    return PVector.mult(PVector.sub(place, pos), 0.001);
  }

  PVector keepInBox()
  {
    float efficiency = 1;
    // efficiency ~ 10 : boids immediately rejected
    // efficiency ~ 1  : boids slowly change direction
    PVector v = new PVector();

    if(pos.x <= width * 0.1)  v.x = efficiency;
    else if(pos.x > width * 0.9)  v.x = -efficiency;
    if(pos.y <= height * 0.1)  v.y = efficiency;
    else if(pos.y > height * 0.9) v.y = -efficiency;

    // show rectangle
    noFill();
    rect(width * 0.1, height * 0.1, width * 0.8, height * 0.8);

    return v;
  }

  void escapeFromMouse()
  {
    float efficiency = 0.03;
    // efficiency ~ 1.0 : boids immediately rejected
    // efficiency ~ .01  : boids slowly change direction
    float distance = dist(mouseX, mouseY, pos.x, pos.y);
    if(distance < aura * 5)
    {
      if(pos.x >= mouseX) vel.x += distance * efficiency;
      else vel.x -= distance * efficiency;
      if(pos.y >= mouseY) vel.y += distance * efficiency;
      else vel.y -= distance * efficiency;
    }
  }

  void updatePosition(Boid[] boids)
  {
    // compute velocity according to rules
    vel.add(goToCenter(boids));
    vel.add(keepDistance(boids));
    vel.add(matchVelocity(boids));
    vel.add(keepInBox());
    limitVelocity(5);
    escapeFromMouse();
    // modify position according to velocity
    pos.add(vel);
  }
}

